/*
 * (C) Copyright 2002 ELTEC Elektronik AG
 * Frank Gottschling <fgottschling@eltec.de>
 *
 * See file CREDITS for list of people who contributed to this
 * project.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of
 * the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.	 See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston,
 * MA 02111-1307 USA
 */

/*
 * cfb_console.c
 *
 * Color Framebuffer Console driver for 8/15/16/24/32 bits per pixel.
 *
 * At the moment only the 8x16 font is tested and the font fore- and
 * background color is limited to black/white/gray colors. The Linux
 * logo can be placed in the upper left corner and additional board
 * information strings (that normaly goes to serial port) can be drawed.
 *
 * The console driver can use the standard PC keyboard interface (i8042)
 * for character input. Character output goes to a memory mapped video
 * framebuffer with little or big-endian organisation.
 * With environment setting 'console=serial' the console i/o can be
 * forced to serial port.

 The driver uses graphic specific defines/parameters/functions:

 (for SMI LynxE graphic chip)

 CONFIG_VIDEO_SMI_LYNXEM - use graphic driver for SMI 710,712,810
 VIDEO_FB_LITTLE_ENDIAN	 - framebuffer organisation default: big endian
 VIDEO_HW_RECTFILL	 - graphic driver supports hardware rectangle fill
 VIDEO_HW_BITBLT	 - graphic driver supports hardware bit blt

 Console Parameters are set by graphic drivers global struct:

 VIDEO_VISIBLE_COLS	     - x resolution
 VIDEO_VISIBLE_ROWS	     - y resolution
 VIDEO_PIXEL_SIZE	     - storage size in byte per pixel
 VIDEO_DATA_FORMAT	     - graphical data format GDF
 VIDEO_FB_ADRS		     - start of video memory

 CONFIG_I8042_KBD	     - AT Keyboard driver for i8042
 VIDEO_KBD_INIT_FCT	     - init function for keyboard
 VIDEO_TSTC_FCT		     - keyboard_tstc function
 VIDEO_GETC_FCT		     - keyboard_getc function

 CONFIG_CONSOLE_CURSOR	     - on/off drawing cursor is done with delay
			       loop in VIDEO_TSTC_FCT (i8042)
 CFG_CONSOLE_BLINK_COUNT     - value for delay loop - blink rate
 CONFIG_CONSOLE_TIME	     - display time/date in upper right corner,
			       needs CFG_CMD_DATE and CONFIG_CONSOLE_CURSOR
 CONFIG_VIDEO_LOGO	     - display Linux Logo in upper left corner
 CONFIG_VIDEO_BMP_LOGO	     - use bmp_logo instead of linux_logo
 CONFIG_CONSOLE_EXTRA_INFO   - display additional board information strings
			       that normaly goes to serial port. This define
			       requires a board specific function:
			       video_drawstring (VIDEO_INFO_X,
						 VIDEO_INFO_Y + i*VIDEO_FONT_HEIGHT,
						 info);
			       that fills a info buffer at i=row.
			       s.a: board/eltec/bab7xx.
CONFIG_VGA_AS_SINGLE_DEVICE  - If set the framebuffer device will be initialised
			       as an output only device. The Keyboard driver
			       will not be set-up. This may be used, if you
			       have none or more than one Keyboard devices
			       (USB Keyboard, AT Keyboard).

CONFIG_VIDEO_SW_CURSOR:	     - Draws a cursor after the last character. No
			       blinking is provided. Uses the macros CURSOR_SET
			       and CURSOR_OFF.
CONFIG_VIDEO_HW_CURSOR:	     - Uses the hardware cursor capability of the
			       graphic chip. Uses the macro CURSOR_SET.
			       ATTENTION: If booting an OS, the display driver
			       must disable the hardware register of the graphic
			       chip. Otherwise a blinking field is displayed
*/

/* this is software sursor shape.it can be modified*/
#define SOFTWARECURSOR	219

#include <rtems.h>

/*****************************************************************************/
/* some Macros								     */
/*****************************************************************************/

#define X800x600

/*
 * Graphic Data Format (GDF) bits for VIDEO_DATA_FORMAT
 */
#define GDF__8BIT_INDEX         0
#define GDF_15BIT_555RGB        1
#define GDF_16BIT_565RGB        2
#define GDF_32BIT_X888RGB       3
#define GDF_24BIT_888RGB        4
#define GDF__8BIT_332RGB        5
#define GDF__1BIT               6
#define GDF__2BIT               7
#define GDF__4BIT               8

#if defined(X800x600)
#define VIDEO_VISIBLE_COLS	800
#define VIDEO_VISIBLE_ROWS	600
#elif defined(X1024x768)
#define VIDEO_VISIBLE_COLS	1024
#define VIDEO_VISIBLE_ROWS	768
#elif defined(X1280x1024)
#define VIDEO_VISIBLE_COLS	1280
#define VIDEO_VISIBLE_ROWS	1024
#elif defined(X640x480)
#define VIDEO_VISIBLE_COLS	640
#define VIDEO_VISIBLE_ROWS	480
#endif

#define VIDEO_PIXEL_SIZE	2
#define VIDEO_DATA_FORMAT	GDF_16BIT_565RGB
#define VIDEO_FB_ADRS		0xc0000000

/*****************************************************************************/
/* Console device							     */
/*****************************************************************************/

#include "video_font.h"

#define VIDEO_COLS		VIDEO_VISIBLE_COLS
#define VIDEO_ROWS		VIDEO_VISIBLE_ROWS
#if  defined(CONFIG_VIDEO_1BPP)
#define VIDEO_SIZE              (VIDEO_ROWS*VIDEO_COLS*VIDEO_PIXEL_SIZE/8)
#define VIDEO_LINE_LEN		(VIDEO_COLS*VIDEO_PIXEL_SIZE/8)
#elif   defined(CONFIG_VIDEO_2BPP)
#define VIDEO_SIZE              (VIDEO_ROWS*VIDEO_COLS*VIDEO_PIXEL_SIZE/4)
#define VIDEO_LINE_LEN		(VIDEO_COLS*VIDEO_PIXEL_SIZE/4)
#elif   defined(CONFIG_VIDEO_4BPP)
#define VIDEO_SIZE              (VIDEO_ROWS*VIDEO_COLS*VIDEO_PIXEL_SIZE/2)
#define VIDEO_LINE_LEN		(VIDEO_COLS*VIDEO_PIXEL_SIZE/2)
#else
#define VIDEO_SIZE		(VIDEO_ROWS*VIDEO_COLS*VIDEO_PIXEL_SIZE)
#define VIDEO_LINE_LEN		(VIDEO_COLS*VIDEO_PIXEL_SIZE)
#endif
#define VIDEO_PIX_BLOCKS	(VIDEO_SIZE >> 2)
#define VIDEO_BURST_LEN		(VIDEO_COLS/8)

#define CONSOLE_ROWS		(VIDEO_ROWS / VIDEO_FONT_HEIGHT)

#define CONSOLE_COLS		(VIDEO_COLS / VIDEO_FONT_WIDTH)
#define CONSOLE_ROW_SIZE	(VIDEO_FONT_HEIGHT * VIDEO_LINE_LEN)
#define CONSOLE_ROW_FIRST	(video_console_address)
#define CONSOLE_ROW_SECOND	(video_console_address + CONSOLE_ROW_SIZE)
#define CONSOLE_ROW_LAST	(video_console_address + CONSOLE_SIZE - CONSOLE_ROW_SIZE)
/* #define CONSOLE_SIZE		(CONSOLE_ROW_SIZE * CONSOLE_ROWS) */
#if  defined(CONFIG_VIDEO_1BPP)
#define CONSOLE_SIZE            (VIDEO_COLS * VIDEO_ROWS * VIDEO_PIXEL_SIZE/8)
#elif   defined(CONFIG_VIDEO_2BPP)
#define CONSOLE_SIZE            (VIDEO_COLS * VIDEO_ROWS * VIDEO_PIXEL_SIZE/4)
#elif   defined(CONFIG_VIDEO_4BPP)
#define CONSOLE_SIZE            (VIDEO_COLS * VIDEO_ROWS * VIDEO_PIXEL_SIZE/2)
#else
#define CONSOLE_SIZE		(VIDEO_COLS * VIDEO_ROWS * VIDEO_PIXEL_SIZE)
#endif

#define CONSOLE_SCROLL_SIZE	(CONSOLE_SIZE - CONSOLE_ROW_SIZE)

/* Macros */
#ifdef	VIDEO_FB_LITTLE_ENDIAN
#define BYTESWAP32(x) 	(x)
#define SWAP16(x)	 ((((x) & 0x00ff) << 8) | ( (x) >> 8))
#define SWAP32(x)	 ((((x) & 0x000000ff) << 24) | (((x) & 0x0000ff00) << 8)|\
                      (((x) & 0x00ff0000) >>  8) | (((x) & 0xff000000) >> 24) )
#define SHORTSWAP32(x)	 ((((x) & 0x000000ff) <<  8) | (((x) & 0x0000ff00) >> 8)|\
                          (((x) & 0x00ff0000) <<  8) | (((x) & 0xff000000) >> 8) )
#else
#define BYTESWAP32(x)    ((((x) & 0x000000ff) <<  24) | (((x) & 0x0000ff00) << 8)|\
                          (((x) & 0x00ff0000) >>  8) | (((x) & 0xff000000) >> 24) )
#define SWAP16(x)	 (x)
#define SWAP32(x)	 (x)
#define SHORTSWAP32(x)	 (x)
#endif

#define CURSOR_SET
#define CURSOR_OFF
#define CONSOLE_BG_COL            0x00
#define CONSOLE_FG_COL            0xff
#define VIDEO_LOGO_HEIGHT	0

#if defined(DEBUG) || defined(DEBUG_CFB_CONSOLE)
#define PRINTD(x)	  printk(x)
#else
#define PRINTD(x)
#endif


static volatile unsigned char *video_fb_address = ( volatile unsigned char * )VIDEO_FB_ADRS;		/* frame buffer address */
static volatile unsigned char *video_console_address = ( volatile unsigned char * )VIDEO_FB_ADRS;	/* console buffer start address */

#ifndef VIDEO_HW_BITBLT
#ifndef MEM_PRINTTO_VIDEO
#define MEM_PRINTTO_VIDEO
static char memfb[CONSOLE_ROWS *CONSOLE_COLS];
#endif
#endif

static int console_col = 0; /* cursor col */
static int console_row = 0; /* cursor row */
unsigned int eorx, fgx, bgx;  /* color pats */

static const char video_font_draw_table2[] =
{
	0x00, 0x03, 0x0c, 0x0f,
	0x30, 0x33, 0x3c, 0x3f,
	0xc0, 0xc3, 0xcc, 0xcf,
	0xf0, 0xf3, 0xfc, 0xff
};
static const char video_font_draw_table4[] =
{
	0x00, 0x0f, 0xf0, 0xff
};
static const int video_font_draw_table8[] =
{
	0x00000000, 0x000000ff, 0x0000ff00, 0x0000ffff,
	0x00ff0000, 0x00ff00ff, 0x00ffff00, 0x00ffffff,
	0xff000000, 0xff0000ff, 0xff00ff00, 0xff00ffff,
	0xffff0000, 0xffff00ff, 0xffffff00, 0xffffffff
};

static const int video_font_draw_table15[] =
{
	0x00000000, 0x00007fff, 0x7fff0000, 0x7fff7fff
};

static const int video_font_draw_table16[] =
{
	0x00000000, 0xffff0000, 0x0000ffff, 0xffffffff
};

static const int video_font_draw_table24[16][3] =
{
	{ 0x00000000, 0x00000000, 0x00000000 },
	{ 0x00000000, 0x00000000, 0x00ffffff },
	{ 0x00000000, 0x0000ffff, 0xff000000 },
	{ 0x00000000, 0x0000ffff, 0xffffffff },
	{ 0x000000ff, 0xffff0000, 0x00000000 },
	{ 0x000000ff, 0xffff0000, 0x00ffffff },
	{ 0x000000ff, 0xffffffff, 0xff000000 },
	{ 0x000000ff, 0xffffffff, 0xffffffff },
	{ 0xffffff00, 0x00000000, 0x00000000 },
	{ 0xffffff00, 0x00000000, 0x00ffffff },
	{ 0xffffff00, 0x0000ffff, 0xff000000 },
	{ 0xffffff00, 0x0000ffff, 0xffffffff },
	{ 0xffffffff, 0xffff0000, 0x00000000 },
	{ 0xffffffff, 0xffff0000, 0x00ffffff },
	{ 0xffffffff, 0xffffffff, 0xff000000 },
	{ 0xffffffff, 0xffffffff, 0xffffffff }
};

static const int video_font_draw_table32[16][4] =
{
	{ 0x00000000, 0x00000000, 0x00000000, 0x00000000 },
	{ 0x00000000, 0x00000000, 0x00000000, 0x00ffffff },
	{ 0x00000000, 0x00000000, 0x00ffffff, 0x00000000 },
	{ 0x00000000, 0x00000000, 0x00ffffff, 0x00ffffff },
	{ 0x00000000, 0x00ffffff, 0x00000000, 0x00000000 },
	{ 0x00000000, 0x00ffffff, 0x00000000, 0x00ffffff },
	{ 0x00000000, 0x00ffffff, 0x00ffffff, 0x00000000 },
	{ 0x00000000, 0x00ffffff, 0x00ffffff, 0x00ffffff },
	{ 0x00ffffff, 0x00000000, 0x00000000, 0x00000000 },
	{ 0x00ffffff, 0x00000000, 0x00000000, 0x00ffffff },
	{ 0x00ffffff, 0x00000000, 0x00ffffff, 0x00000000 },
	{ 0x00ffffff, 0x00000000, 0x00ffffff, 0x00ffffff },
	{ 0x00ffffff, 0x00ffffff, 0x00000000, 0x00000000 },
	{ 0x00ffffff, 0x00ffffff, 0x00000000, 0x00ffffff },
	{ 0x00ffffff, 0x00ffffff, 0x00ffffff, 0x00000000 },
	{ 0x00ffffff, 0x00ffffff, 0x00ffffff, 0x00ffffff }
};


static void memsetl ( int *p, int c, int v );
void video_putchar ( int xx, int yy, unsigned char c );
static void memcpyl ( int *d, int *s, int c );

/******************************************************************************/
void video_drawchars ( int xx, int yy, unsigned char *s, int count )
{
	unsigned char *cdat;
	volatile unsigned char *dest0, *dest;
	int rows, offset, c;
	int i;
	offset = yy * VIDEO_LINE_LEN + xx * VIDEO_PIXEL_SIZE;
	dest0 = video_fb_address + offset;

	switch ( VIDEO_DATA_FORMAT )
	{
		case GDF__1BIT:
			while ( count-- )
			{
				c = *s;
				cdat = video_fontdata + c * VIDEO_FONT_HEIGHT;

				for ( i = 0; i < VIDEO_FONT_HEIGHT; i++ )
				{
					offset = ( ( yy + i ) * ( VIDEO_VISIBLE_COLS ) + xx ) / 8;
					*( unsigned char * )( video_fb_address + offset ) = ( ( unsigned char * )cdat )[i];
				}

				s++;
			}

			break;

		case GDF__2BIT:
			while ( count-- )
			{
				c = *s;
				cdat = video_fontdata + c * VIDEO_FONT_HEIGHT;

				for ( i = 0; i < VIDEO_FONT_HEIGHT; i++ )
				{
					unsigned char bits = *cdat++;
					offset = ( ( yy + i ) * ( VIDEO_VISIBLE_COLS ) + xx ) / 4;
					dest = video_fb_address + offset;
					dest[0] = video_font_draw_table2[bits >> 4] ;
					dest[1] = video_font_draw_table2[bits & 15] ;
				}

				s++;
			}

			break;

		case GDF__4BIT:
			while ( count-- )
			{
				c = *s;
				cdat = video_fontdata + c * VIDEO_FONT_HEIGHT;

				for ( i = 0; i < VIDEO_FONT_HEIGHT; i++ )
				{
					unsigned char bits = *cdat++;
					offset = ( ( yy + i ) * ( VIDEO_VISIBLE_COLS ) + xx ) / 2;
					dest = video_fb_address + offset;
					dest[0] = video_font_draw_table4[bits >> 6] ;
					dest[1] = video_font_draw_table4[bits >> 4 & 3] ;
					dest[2] = video_font_draw_table4[bits >> 2 & 3] ;
					dest[3] = video_font_draw_table4[bits & 3] ;
				}

				s++;
			}

			break;

		case GDF__8BIT_INDEX:
		case GDF__8BIT_332RGB:
			while ( count-- )
			{
				c = *s;
				cdat = video_fontdata + c * VIDEO_FONT_HEIGHT;

				for ( rows = VIDEO_FONT_HEIGHT, dest = dest0;
				        rows--;
				        dest += VIDEO_LINE_LEN )
				{
					unsigned char bits = *cdat++;
					( ( volatile unsigned int * ) dest )[0] = BYTESWAP32( ( video_font_draw_table8[bits >> 4] & eorx ) ^ bgx );
					( ( volatile unsigned int * ) dest )[1] = BYTESWAP32( ( video_font_draw_table8[bits & 15] & eorx ) ^ bgx );
				}

				dest0 += VIDEO_FONT_WIDTH * VIDEO_PIXEL_SIZE;
				s++;
			}

			break;

		case GDF_15BIT_555RGB:
			while ( count-- )
			{
				c = *s;
				cdat = video_fontdata + c * VIDEO_FONT_HEIGHT;

				for ( rows = VIDEO_FONT_HEIGHT, dest = dest0;
				        rows--;
				        dest += VIDEO_LINE_LEN )
				{
					unsigned char bits = *cdat++;
					( ( volatile unsigned int * ) dest )[0] = SHORTSWAP32 ( ( video_font_draw_table15 [bits >> 6] & eorx ) ^ bgx );
					( ( volatile unsigned int * ) dest )[1] = SHORTSWAP32 ( ( video_font_draw_table15 [bits >> 4 & 3] & eorx ) ^ bgx );
					( ( volatile unsigned int * ) dest )[2] = SHORTSWAP32 ( ( video_font_draw_table15 [bits >> 2 & 3] & eorx ) ^ bgx );
					( ( volatile unsigned int * ) dest )[3] = SHORTSWAP32 ( ( video_font_draw_table15 [bits & 3] & eorx ) ^ bgx );
				}

				dest0 += VIDEO_FONT_WIDTH * VIDEO_PIXEL_SIZE;
				s++;
			}

			break;

			/*here*/
		case GDF_16BIT_565RGB:
			while ( count-- )
			{
				c = *s;
				cdat = video_fontdata + c * VIDEO_FONT_HEIGHT;

				for ( rows = VIDEO_FONT_HEIGHT, dest = dest0;
				        rows--;
				        dest += VIDEO_LINE_LEN )
				{
					unsigned char bits = *cdat++;
					( ( volatile unsigned int * ) dest )[0] = SHORTSWAP32 ( ( video_font_draw_table16 [bits >> 6] & eorx ) ^ bgx );
					( ( volatile unsigned int * ) dest )[1] = SHORTSWAP32 ( ( video_font_draw_table16 [bits >> 4 & 3] & eorx ) ^ bgx );
					( ( volatile unsigned int * ) dest )[2] = SHORTSWAP32 ( ( video_font_draw_table16 [bits >> 2 & 3] & eorx ) ^ bgx );
					( ( volatile unsigned int * ) dest )[3] = SHORTSWAP32 ( ( video_font_draw_table16 [bits & 3] & eorx ) ^ bgx );
				}

				dest0 += VIDEO_FONT_WIDTH * VIDEO_PIXEL_SIZE;
				s++;
			}

			break;

		case GDF_32BIT_X888RGB:
			while ( count-- )
			{
				c = *s;
				cdat = video_fontdata + c * VIDEO_FONT_HEIGHT;

				for ( rows = VIDEO_FONT_HEIGHT, dest = dest0;
				        rows--;
				        dest += VIDEO_LINE_LEN )
				{
					unsigned char bits = *cdat++;
					( ( volatile unsigned int * ) dest )[0] = SWAP32 ( ( video_font_draw_table32 [bits >> 4][0] & eorx ) ^ bgx );
					( ( volatile unsigned int * ) dest )[1] = SWAP32 ( ( video_font_draw_table32 [bits >> 4][1] & eorx ) ^ bgx );
					( ( volatile unsigned int * ) dest )[2] = SWAP32 ( ( video_font_draw_table32 [bits >> 4][2] & eorx ) ^ bgx );
					( ( volatile unsigned int * ) dest )[3] = SWAP32 ( ( video_font_draw_table32 [bits >> 4][3] & eorx ) ^ bgx );
					( ( volatile unsigned int * ) dest )[4] = SWAP32 ( ( video_font_draw_table32 [bits & 15][0] & eorx ) ^ bgx );
					( ( volatile unsigned int * ) dest )[5] = SWAP32 ( ( video_font_draw_table32 [bits & 15][1] & eorx ) ^ bgx );
					( ( volatile unsigned int * ) dest )[6] = SWAP32 ( ( video_font_draw_table32 [bits & 15][2] & eorx ) ^ bgx );
					( ( volatile unsigned int * ) dest )[7] = SWAP32 ( ( video_font_draw_table32 [bits & 15][3] & eorx ) ^ bgx );
				}

				dest0 += VIDEO_FONT_WIDTH * VIDEO_PIXEL_SIZE;
				s++;
			}

			break;

		case GDF_24BIT_888RGB:
			while ( count-- )
			{
				c = *s;
				cdat = video_fontdata + c * VIDEO_FONT_HEIGHT;

				for ( rows = VIDEO_FONT_HEIGHT, dest = dest0;
				        rows--;
				        dest += VIDEO_LINE_LEN )
				{
					unsigned char bits = *cdat++;
					( ( volatile unsigned int * ) dest )[0] = ( video_font_draw_table24[bits >> 4][0] & eorx ) ^ bgx;
					( ( volatile unsigned int * ) dest )[1] = ( video_font_draw_table24[bits >> 4][1] & eorx ) ^ bgx;
					( ( volatile unsigned int * ) dest )[2] = ( video_font_draw_table24[bits >> 4][2] & eorx ) ^ bgx;
					( ( volatile unsigned int * ) dest )[3] = ( video_font_draw_table24[bits & 15][0] & eorx ) ^ bgx;
					( ( volatile unsigned int * ) dest )[4] = ( video_font_draw_table24[bits & 15][1] & eorx ) ^ bgx;
					( ( volatile unsigned int * ) dest )[5] = ( video_font_draw_table24[bits & 15][2] & eorx ) ^ bgx;
				}

				dest0 += VIDEO_FONT_WIDTH * VIDEO_PIXEL_SIZE;
				s++;
			}

			break;
	}
}

/*****************************************************************************/

void video_drawstring ( int xx, int yy, unsigned char *s )
{
	video_drawchars ( xx, yy, s, strlen ( ( char * )s ) );
}

/*****************************************************************************/

#ifndef VIDEO_HW_BITBLT
void video_drawsline( char *str, int rows, int cols )
{
	int xx, yy;
	int pos;

	for ( yy = 1; yy < rows; yy++ )
	{
		pos = yy * cols;

		for ( xx = 0; xx < cols; xx++ )
		{
			if ( str[pos + xx] != str[pos - cols + xx] )
			{
				video_putchar( xx * VIDEO_FONT_WIDTH, ( yy - 1 ) * VIDEO_FONT_HEIGHT, str[pos + xx] );
			}
		}
	}

	memcpyl ( ( int * )str, ( int * )&str[cols], ( rows - 1 ) *cols >> 2 );
	memsetl ( ( int * )&str[( rows - 1 ) * cols], cols >> 2, CONSOLE_BG_COL );
}

#endif

void video_putchar ( int xx, int yy, unsigned char c )
{
	video_drawchars ( xx, yy + VIDEO_LOGO_HEIGHT, &c, 1 );
}
#ifdef INTERFACE_3A780E
void video_putchar1( int xx, int yy, unsigned char c )
{
	video_drawchars ( xx, yy, &c, 1 );
}
#endif

/*****************************************************************************/

#if 1

static void memsetl ( int *p, int c, int v )
{
	while ( c-- )
	{
		*( p++ ) = v;
	}
}
#endif

/*****************************************************************************/

#ifndef VIDEO_HW_BITBLT
static void memcpyl ( int *d, int *s, int c )
{
	while ( c-- )
	{
		*( d++ ) = *( s++ );
	}
}
#endif

/*****************************************************************************/

static void console_scrollup ( void )
{
	/* copy up rows ignoring the first one */
#if defined(MEM_PRINTTO_VIDEO)
	video_drawsline( memfb, CONSOLE_ROWS, CONSOLE_COLS );
#else
#ifdef __mips__
#define UNCACHED_TO_CACHED(x) PHYS_TO_CACHED(UNCACHED_TO_PHYS(x))

	if ( CONSOLE_ROW_FIRST < 0xb0000000 && CONSOLE_ROW_FIRST >= 0xa0000000 )
	{
		memcpyl ( UNCACHED_TO_CACHED( CONSOLE_ROW_FIRST ), UNCACHED_TO_CACHED( CONSOLE_ROW_SECOND ),
		          CONSOLE_SCROLL_SIZE >> 2 );
		pci_sync_cache( 0, ( vm_offset_t )UNCACHED_TO_CACHED( CONSOLE_ROW_FIRST ), CONSOLE_SCROLL_SIZE >> 2, SYNC_W );
	}

	else
#endif
		memcpyl ( CONSOLE_ROW_FIRST, CONSOLE_ROW_SECOND,
		          CONSOLE_SCROLL_SIZE >> 2 );

#endif
	/* clear the last one */
#ifdef VIDEO_HW_RECTFILL
#ifdef RADEON7000
	video_hw_rectfill ( VIDEO_PIXEL_SIZE,	/* bytes per pixel */
	                    0,	/* dest pos x */
	                    VIDEO_VISIBLE_ROWS - VIDEO_FONT_HEIGHT,	/* dest pos y */
	                    VIDEO_VISIBLE_COLS,	/* frame width */
	                    VIDEO_FONT_HEIGHT,	/* frame height */
	                    CONSOLE_BG_COL	/* fill color */
	                  );
#elif (defined(SMI502)||defined(SMI712))
	AutodeFillRectModify( CONSOLE_BG_COL );
#ifdef X800x480
	memsetl ( CONSOLE_ROW_LAST, CONSOLE_ROW_SIZE >> 2, CONSOLE_BG_COL );
#endif
#else
	memsetl ( CONSOLE_ROW_LAST, CONSOLE_ROW_SIZE >> 2, CONSOLE_BG_COL );
#endif
#elif defined(X800x600)
	memsetl ( ( int * )( CONSOLE_ROW_LAST - CONSOLE_ROW_SIZE / 2 ), CONSOLE_ROW_SIZE >> 2, CONSOLE_BG_COL );
#else
	memsetl ( ( int * )( CONSOLE_ROW_LAST - CONSOLE_ROW_SIZE / 2 ), CONSOLE_ROW_SIZE >> 2, CONSOLE_BG_COL );
#endif
}

/*****************************************************************************/

static void console_back ( void )
{
	console_col--;

	if ( console_col < 0 )
	{
		console_col = CONSOLE_COLS - 1;
		console_row--;

		if ( console_row < 0 )
		{
			console_row = 0;
		}
	}
}

/*****************************************************************************/

static void console_newline ( void )
{
	console_row++;
	console_col = 0;

	/* Check if we need to scroll the terminal */
	if ( console_row >= CONSOLE_ROWS )
	{
		/* Scroll everything up */
		console_scrollup ();
		/* Decrement row number */
		console_row--;
	}
}

/*****************************************************************************/

void video_putc ( const char c )
{
	switch ( c )
	{
		case 13:		/* ignore */
			CURSOR_OFF
			break;

		case '\n':		/* next line */
			CURSOR_OFF
			console_newline ();
			break;

		case 9:		/* tab 8 */
			CURSOR_OFF console_col |= 0x0008;
			console_col &= ~0x0007;

			if ( console_col >= CONSOLE_COLS )
			{
				console_newline ();
			}

			break;

		case 8:		/* backspace */
			CURSOR_OFF
			console_back ();
			break;

		default
				:		/* draw the char */
			video_putchar ( console_col * VIDEO_FONT_WIDTH,
			                console_row * VIDEO_FONT_HEIGHT,
			                c );
#ifdef MEM_PRINTTO_VIDEO
			memfb[console_row * CONSOLE_COLS + console_col] = c;
#endif
			console_col++;

			/* check for newline */
			if ( console_col >= CONSOLE_COLS )
			{
				console_newline ();
			}
	}

	CURSOR_SET
}


/*****************************************************************************/

void video_cls_all( void )
{
	memsetl ( ( int * )video_fb_address,
	          CONSOLE_SIZE,
	          CONSOLE_BG_COL );
}

void video_set_background( unsigned char r, unsigned char g, unsigned char b )
{
	int cnt = CONSOLE_SIZE - VIDEO_LOGO_HEIGHT * VIDEO_LINE_LEN;
	volatile unsigned short *p = ( volatile unsigned short * )( video_fb_address + VIDEO_LOGO_HEIGHT * VIDEO_LINE_LEN );
	unsigned short color = SWAP16 ( ( unsigned short ) ( ( ( r >> 3 ) << 11 ) | ( ( g >> 2 ) << 5 ) | ( b >> 3 ) ) );

	while ( cnt > 0 )
	{
		*( p++ ) = color;
		cnt -= 2;
	}
}

static int record = 1;
/*80*24*/
#if defined(X640x480)
char console_buffer[2][31][81] = {{{32}}}; /*80*30->640x480*/
#elif defined(X800x480)
char console_buffer[2][37][101] = {32};
#elif defined(X800x600)
char console_buffer[2][37][101] = {{{32}}};
#elif defined(X1024x768)
char console_buffer[2][49][129] = {32}; /*128*48->1024x768*/
#elif defined(X1280x1024)
char console_buffer[2][65][161] = {32}; /*128*48->1024x768*/
#elif defined(X320x240)
char console_buffer[2][16][41] = {32}; /*40*15->320x240*/
#else
char console_buffer[2][31][81] = {32}; /*80*30->640x480*/
#endif

void video_console_print( int console_col, int console_row, unsigned char *s )
{
	int count = strlen ( ( char * )s );

	while ( count-- )
	{
		video_putchar ( console_col * VIDEO_FONT_WIDTH,
		                console_row * VIDEO_FONT_HEIGHT,
		                *s );

		if ( record )
		{
			console_buffer[0][console_row][console_col] = *s;
			console_buffer[1][console_row][console_col] = ( char )bgx;
		}

		console_col++;
		s++;
	}
}
void begin_record( void )
{
	record = 1;
}

void stop_record( void )
{
	record = 0;
}


char video_get_console_char( int console_col, int console_row )
{
	return console_buffer[0][console_row][console_col];
}

char video_get_console_bgcolor( int console_col, int console_row )
{
	return console_buffer[1][console_row][console_col];
}

void video_set_bg( unsigned char r, unsigned char g, unsigned char b )
{
	bgx = SWAP16 ( ( unsigned short ) ( ( ( r >> 3 ) << 11 ) | ( ( g >> 2 ) << 5 ) | ( b >> 3 ) ) );
	bgx |= ( bgx << 16 );
	fgx =  SWAP16 ( ( unsigned short ) ( ( ( 128 >> 3 ) << 11 ) | ( ( 128 >> 2 ) << 5 ) | ( 128 >> 3 ) ) );
	fgx |= ( fgx << 16 );
	eorx = fgx ^ bgx;
}


unsigned short pallete[16] =
{
	SWAP16 ( ( unsigned short ) ( ( ( 0   >> 3 ) << 11 ) | ( ( 0   >> 2 ) << 5 ) | ( 0   >> 3 ) ) ),
	SWAP16 ( ( unsigned short ) ( ( ( 0   >> 3 ) << 11 ) | ( ( 0   >> 2 ) << 5 ) | ( 128 >> 3 ) ) ),
	SWAP16 ( ( unsigned short ) ( ( ( 0   >> 3 ) << 11 ) | ( ( 128 >> 2 ) << 5 ) | ( 0   >> 3 ) ) ),
	SWAP16 ( ( unsigned short ) ( ( ( 0   >> 3 ) << 11 ) | ( ( 128 >> 2 ) << 5 ) | ( 128 >> 3 ) ) ),
	SWAP16 ( ( unsigned short ) ( ( ( 128 >> 3 ) << 11 ) | ( ( 0   >> 2 ) << 5 ) | ( 0   >> 3 ) ) ),
	SWAP16 ( ( unsigned short ) ( ( ( 128 >> 3 ) << 11 ) | ( ( 0   >> 2 ) << 5 ) | ( 128 >> 3 ) ) ),
	SWAP16 ( ( unsigned short ) ( ( ( 128 >> 3 ) << 11 ) | ( ( 128 >> 2 ) << 5 ) | ( 0   >> 3 ) ) ),
	SWAP16 ( ( unsigned short ) ( ( ( 0xc0 >> 3 ) << 11 ) | ( ( 0xc0 >> 2 ) << 5 ) | ( 0xc0 >> 3 ) ) ),
	SWAP16 ( ( unsigned short ) ( ( ( 128 >> 3 ) << 11 ) | ( ( 128 >> 2 ) << 5 ) | ( 128 >> 3 ) ) ),
	SWAP16 ( ( unsigned short ) ( ( ( 0   >> 3 ) << 11 ) | ( ( 0   >> 2 ) << 5 ) | ( 255 >> 3 ) ) ),
	SWAP16 ( ( unsigned short ) ( ( ( 0   >> 3 ) << 11 ) | ( ( 255 >> 2 ) << 5 ) | ( 0   >> 3 ) ) ),
	SWAP16 ( ( unsigned short ) ( ( ( 0   >> 3 ) << 11 ) | ( ( 255 >> 2 ) << 5 ) | ( 255 >> 3 ) ) ),
	SWAP16 ( ( unsigned short ) ( ( ( 255 >> 3 ) << 11 ) | ( ( 0   >> 2 ) << 5 ) | ( 0   >> 3 ) ) ),
	SWAP16 ( ( unsigned short ) ( ( ( 255 >> 3 ) << 11 ) | ( ( 0   >> 2 ) << 5 ) | ( 255 >> 3 ) ) ),
	SWAP16 ( ( unsigned short ) ( ( ( 255 >> 3 ) << 11 ) | ( ( 255 >> 2 ) << 5 ) | ( 0   >> 3 ) ) ),
	SWAP16 ( ( unsigned short ) ( ( ( 255 >> 3 ) << 11 ) | ( ( 255 >> 2 ) << 5 ) | ( 255 >> 3 ) ) )
};

#ifdef INTERFACE_3A780E
void video_set_color( unsigned char color )
{
#ifndef FB_MENU_NOCLOLOR
	bgx = pallete[color >> 4];
	bgx |= ( bgx << 16 );
	fgx =  pallete[color & 0xf];
	fgx |= ( fgx << 16 );
	eorx = fgx ^ bgx;
#endif
}
#endif
static void __cprint( int y, int x, int width, char color, const char *buf )
{
#ifndef FB_MENU_NOCLOLOR
	bgx = pallete[color >> 4];
	bgx |= ( bgx << 16 );
	fgx =  pallete[color & 0xf];
	fgx |= ( fgx << 16 );
	eorx = fgx ^ bgx;
#endif
	video_console_print( x, y, ( unsigned char * )buf );
#ifndef FB_MENU_NOCLOLOR
	bgx = 0;
	fgx =  SWAP16 ( ( unsigned short ) ( ( ( 128 >> 3 ) << 11 ) | ( ( 128 >> 2 ) << 5 ) | ( 128 >> 3 ) ) );
	fgx |= ( fgx << 16 );
	eorx = fgx ^ bgx;
#endif
}

void cprintfb( int y, int x, int width, char color, const char *text )
{
	int i, l;
	char buf[200];

	if ( width == 0 && text )
	{
		width = strlen( text );
	}

	if ( text && ( l = strlen( text ) ) )
	{
		memcpy( buf, text, ( width && l > width ) ? width : l );
	}

	else
	{
		l = 0;
	}

	if ( l < width )
		for ( i = 0; i < width - l; i++ )
		{
			buf[l + i] = 0x20;
		}

	if ( width )
	{
		buf[width] = 0;
		__cprint( y, x, width, color, buf );
	}
}

void set_cursor_fb( unsigned char x, unsigned char y )
{
	console_col = x;
	console_row = y;
}

int fb_init( void )
{
	unsigned char color8;

	/* AppInit drawing pats */
	switch ( VIDEO_DATA_FORMAT )
	{
#if 0

		case GDF__8BIT_INDEX:
			video_set_lut ( 0x01, CONSOLE_FG_COL, CONSOLE_FG_COL, CONSOLE_FG_COL );
			video_set_lut ( 0x00, CONSOLE_BG_COL, CONSOLE_BG_COL, CONSOLE_BG_COL );
			fgx = 0x01010101;
			bgx = 0x00000000;
			break;
#endif

		case GDF__1BIT:
			fgx = 1;
			bgx = 0;
			break;

		case GDF__2BIT:
			fgx = 3;
			bgx = 0;
			break;

		case GDF__4BIT:
			fgx = 0xf;
			bgx = 0;
			break;

		case GDF__8BIT_332RGB:
			color8 = ( ( CONSOLE_FG_COL & 0xe0 ) |
			           ( ( CONSOLE_FG_COL >> 3 ) & 0x1c ) | CONSOLE_FG_COL >> 6 );
			fgx = ( color8 << 24 ) | ( color8 << 16 ) | ( color8 << 8 ) | color8;
			color8 = ( ( CONSOLE_BG_COL & 0xe0 ) |
			           ( ( CONSOLE_BG_COL >> 3 ) & 0x1c ) | CONSOLE_BG_COL >> 6 );
			bgx = ( color8 << 24 ) | ( color8 << 16 ) | ( color8 << 8 ) | color8;
			break;

		case GDF_15BIT_555RGB:
			fgx = ( ( ( CONSOLE_FG_COL >> 3 ) << 26 ) |
			        ( ( CONSOLE_FG_COL >> 3 ) << 21 ) | ( ( CONSOLE_FG_COL >> 3 ) << 16 ) |
			        ( ( CONSOLE_FG_COL >> 3 ) << 10 ) | ( ( CONSOLE_FG_COL >> 3 ) << 5 ) |
			        ( CONSOLE_FG_COL >> 3 ) );
			bgx = ( ( ( CONSOLE_BG_COL >> 3 ) << 26 ) |
			        ( ( CONSOLE_BG_COL >> 3 ) << 21 ) | ( ( CONSOLE_BG_COL >> 3 ) << 16 ) |
			        ( ( CONSOLE_BG_COL >> 3 ) << 10 ) | ( ( CONSOLE_BG_COL >> 3 ) << 5 ) |
			        ( CONSOLE_BG_COL >> 3 ) );
			break;

		case GDF_16BIT_565RGB:
			fgx = ( ( ( CONSOLE_FG_COL >> 3 ) << 27 ) |
			        ( ( CONSOLE_FG_COL >> 2 ) << 21 ) | ( ( CONSOLE_FG_COL >> 3 ) << 16 ) |
			        ( ( CONSOLE_FG_COL >> 3 ) << 11 ) | ( ( CONSOLE_FG_COL >> 2 ) << 5 ) |
			        ( CONSOLE_FG_COL >> 3 ) );
			bgx = ( ( ( CONSOLE_BG_COL >> 3 ) << 27 ) |
			        ( ( CONSOLE_BG_COL >> 2 ) << 21 ) | ( ( CONSOLE_BG_COL >> 3 ) << 16 ) |
			        ( ( CONSOLE_BG_COL >> 3 ) << 11 ) | ( ( CONSOLE_BG_COL >> 2 ) << 5 ) |
			        ( CONSOLE_BG_COL >> 3 ) );
			break;

		case GDF_32BIT_X888RGB:
			fgx = ( CONSOLE_FG_COL << 16 ) | ( CONSOLE_FG_COL << 8 ) | CONSOLE_FG_COL;
			bgx = ( CONSOLE_BG_COL << 16 ) | ( CONSOLE_BG_COL << 8 ) | CONSOLE_BG_COL;
			break;

		case GDF_24BIT_888RGB:
			fgx = ( CONSOLE_FG_COL << 24 ) | ( CONSOLE_FG_COL << 16 ) |
			      ( CONSOLE_FG_COL << 8 ) | CONSOLE_FG_COL;
			bgx = ( CONSOLE_BG_COL << 24 ) | ( CONSOLE_BG_COL << 16 ) |
			      ( CONSOLE_BG_COL << 8 ) | CONSOLE_BG_COL;
			break;
	}

	eorx = fgx ^ bgx;
	memsetl ( ( int * )video_fb_address, VIDEO_COLS * VIDEO_ROWS * VIDEO_PIXEL_SIZE / 4, CONSOLE_BG_COL );
	/* Initialize the console */
	console_col = 0;
	console_row = 0;
	memset ( console_buffer, ' ', sizeof console_buffer );
#if defined(MEM_PRINTTO_VIDEO)
	memset ( memfb, 0, CONSOLE_ROWS * CONSOLE_COLS );
#endif
	return 0;
}

